from plain.urls import Router, path

from . import views


class LoginlinkRouter(Router):
    namespace = "loginlink"
    urls = [
        path("sent/", views.LoginLinkSentView.as_view(), name="sent"),
        path("failed/", views.LoginLinkFailedView.as_view(), name="failed"),
        path("token/<str:token>/", views.LoginLinkLoginView.as_view(), name="login"),
    ]



from __future__ import annotations

from typing import Any

from plain.auth import login, logout
from plain.auth.views import AuthView
from plain.forms import BaseForm
from plain.http import RedirectResponse, Response
from plain.runtime import settings
from plain.urls import reverse, reverse_lazy
from plain.views import FormView, TemplateView, View

from .forms import LoginLinkForm
from .links import (
    LoginLinkChanged,
    LoginLinkExpired,
    LoginLinkInvalid,
    get_link_token_user,
)


class LoginLinkFormView(AuthView, FormView):
    form_class = LoginLinkForm
    success_url = reverse_lazy("loginlink:sent")

    def get(self) -> Response:
        # Redirect if the user is already logged in
        if self.user:
            form = self.get_form()
            return RedirectResponse(self.get_success_url(form))

        return super().get()

    def form_valid(self, form: LoginLinkForm) -> Response:  # type: ignore[override]
        form.maybe_send_link(self.request)
        return super().form_valid(form)

    def get_success_url(self, form: BaseForm) -> str:
        if next_url := form.cleaned_data.get("next"):
            # Keep the next URL in the query string so the sent
            # view can redirect to it if reloaded and logged in already.
            return f"{self.success_url}?next={next_url}"
        else:
            return self.success_url


class LoginLinkSentView(AuthView, TemplateView):
    template_name = "loginlink/sent.html"

    def get(self) -> Response:
        # Redirect if the user is already logged in
        if self.user:
            next_url = self.request.query_params.get("next", "/")
            return RedirectResponse(next_url)

        return super().get()


class LoginLinkFailedView(TemplateView):
    template_name = "loginlink/failed.html"

    def get_template_context(self) -> dict[str, Any]:
        context = super().get_template_context()
        context["error"] = self.request.query_params.get("error")
        context["login_url"] = reverse(settings.AUTH_LOGIN_URL)
        return context


class LoginLinkLoginView(AuthView, View):
    success_url = "/"

    def get(self) -> Response:
        # If they're logged in, log them out and process the link again
        if self.user:
            logout(self.request)

        token = self.url_kwargs["token"]

        try:
            user = get_link_token_user(token)
        except LoginLinkExpired:
            return RedirectResponse(reverse("loginlink:failed") + "?error=expired")
        except LoginLinkInvalid:
            return RedirectResponse(reverse("loginlink:failed") + "?error=invalid")
        except LoginLinkChanged:
            return RedirectResponse(reverse("loginlink:failed") + "?error=changed")

        login(self.request, user)

        if next_url := self.request.query_params.get("next"):
            return RedirectResponse(next_url)

        return RedirectResponse(self.success_url)



from __future__ import annotations

import time
import zlib
from typing import Any

from plain.signing import (
    JSONSerializer,
    SignatureExpired,
    Signer,
    b62_decode,
    b62_encode,
    b64_decode,
    b64_encode,
)


class ExpiringSigner:
    """A signer with an embedded expiration (vs max age unsign).

    Uses composition rather than inheritance since the interface
    intentionally differs from Signer (requires expires_in parameter).
    """

    def __init__(
        self,
        *,
        key: str | None = None,
        sep: str = ":",
        salt: str | None = None,
        algorithm: str = "sha256",
        fallback_keys: list[str] | None = None,
    ) -> None:
        # Compute default salt here to preserve backwards compatibility.
        # When ExpiringSigner inherited from Signer, the default salt was
        # "plain.loginlink.signing.ExpiringSigner". Now that we use composition,
        # we must set it explicitly rather than letting Signer compute its own.
        if salt is None:
            salt = f"{self.__class__.__module__}.{self.__class__.__name__}"
        self._signer = Signer(
            key=key,
            sep=sep,
            salt=salt,
            algorithm=algorithm,
            fallback_keys=fallback_keys,
        )

    @property
    def sep(self) -> str:
        return self._signer.sep

    def sign(self, value: str, expires_in: int) -> str:
        timestamp = b62_encode(int(time.time() + expires_in))
        value = f"{value}{self.sep}{timestamp}"
        return self._signer.sign(value)

    def unsign(self, signed_value: str) -> str:
        """
        Retrieve original value and check the expiration hasn't passed.
        """
        result = self._signer.unsign(signed_value)
        value, timestamp = result.rsplit(self.sep, 1)
        ts = b62_decode(timestamp)
        if ts < time.time():
            raise SignatureExpired("Signature expired")
        return value

    def sign_object(
        self,
        obj: Any,
        *,
        expires_in: int,
        serializer: type = JSONSerializer,
        compress: bool = False,
    ) -> str:
        """
        Return URL-safe, hmac signed base64 compressed JSON string.

        If compress is True (not the default), check if compressing using zlib
        can save some space. Prepend a '.' to signify compression. This is
        included in the signature, to protect against zip bombs.

        The serializer is expected to return a bytestring.
        """
        data = serializer().dumps(obj)
        # Flag for if it's been compressed or not.
        is_compressed = False

        if compress:
            # Avoid zlib dependency unless compress is being used.
            compressed = zlib.compress(data)
            if len(compressed) < (len(data) - 1):
                data = compressed
                is_compressed = True
        base64d = b64_encode(data).decode()
        if is_compressed:
            base64d = "." + base64d
        return self.sign(base64d, expires_in)

    def unsign_object(self, signed_obj: str, serializer: type = JSONSerializer) -> Any:
        # Signer.unsign() returns str but base64 and zlib compression operate
        # on bytes.
        base64d = self.unsign(signed_obj).encode()
        decompress = base64d[:1] == b"."
        if decompress:
            # It's compressed; uncompress it first.
            base64d = base64d[1:]
        data = b64_decode(base64d)
        if decompress:
            data = zlib.decompress(data)
        return serializer().loads(data)


def dumps(
    obj: Any,
    *,
    expires_in: int,
    key: str | None = None,
    salt: str = "plain.loginlink",
    serializer: type = JSONSerializer,
    compress: bool = False,
) -> str:
    """
    Return URL-safe, hmac signed base64 compressed JSON string. If key is
    None, use settings.SECRET_KEY instead. The hmac algorithm is the default
    Signer algorithm.

    If compress is True (not the default), check if compressing using zlib can
    save some space. Prepend a '.' to signify compression. This is included
    in the signature, to protect against zip bombs.

    Salt can be used to namespace the hash, so that a signed string is
    only valid for a given namespace. Leaving this at the default
    value or re-using a salt value across different parts of your
    application without good cause is a security risk.

    The serializer is expected to return a bytestring.
    """
    return ExpiringSigner(key=key, salt=salt).sign_object(
        obj, expires_in=expires_in, serializer=serializer, compress=compress
    )


def loads(
    s: str,
    *,
    key: str | None = None,
    salt: str = "plain.loginlink",
    serializer: type = JSONSerializer,
    fallback_keys: list[str] | None = None,
) -> Any:
    """
    Reverse of dumps(), raise BadSignature if signature fails.

    The serializer is expected to accept a bytestring.
    """
    return ExpiringSigner(
        key=key, salt=salt, fallback_keys=fallback_keys
    ).unsign_object(
        s,
        serializer=serializer,
    )



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from plain import forms
from plain.auth import get_user_model
from plain.email import TemplateEmail

from .links import generate_link_url

if TYPE_CHECKING:
    from plain.http import Request


class LoginLinkForm(forms.Form):
    email = forms.EmailField()
    next = forms.CharField(required=False)

    def maybe_send_link(
        self, request: Request, expires_in: int = 60 * 60
    ) -> int | None:
        user_model = get_user_model()
        email = self.cleaned_data["email"]
        try:
            user = user_model.query.get(email__iexact=email)
        except user_model.DoesNotExist:
            user = None

        if user:
            url = generate_link_url(
                request=request, user=user, email=email, expires_in=expires_in
            )

            if next_url := self.cleaned_data.get("next"):
                url += f"?next={next_url}"

            email = self.get_template_email(
                email=email,
                context={"user": user, "url": url, "expires_in": expires_in},
            )
            return email.send()

        return None

    def get_template_email(
        self, *, email: str, context: dict[str, Any]
    ) -> TemplateEmail:
        return TemplateEmail(
            template="loginlink",
            to=[email],
            context=context,
        )
